//===- ProfileInfoLoad.cpp - Load profile information from disk -----------===//
//
//                      The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// The ProfileInfoLoader class is used to load and represent profiling
// information read in from the dump file.
//
//===----------------------------------------------------------------------===//

#include "llvm/Analysis/ProfileInfoLoader.h"
#include "llvm/Analysis/ProfileInfoTypes.h"
#include "llvm/Module.h"
#include "llvm/InstrTypes.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Debug.h"
#include <cstdio>
#include <cstdlib>
#include <map>
using namespace llvm;

// ByteSwap - Byteswap 'Var' if 'Really' is true.
//
static inline unsigned ByteSwap(unsigned Var, bool Really) {
  if (!Really) return Var;
  return ((Var & (255U<< 0U)) << 24U) |
         ((Var & (255U<< 8U)) <<  8U) |
         ((Var & (255U<<16U)) >>  8U) |
         ((Var & (255U<<24U)) >> 24U);
}

static unsigned AddCounts(unsigned A, unsigned B) {
  // If either value is undefined, use the other.
  if (A == ProfileInfoLoader::Uncounted) return B;
  if (B == ProfileInfoLoader::Uncounted) return A;
  return A + B;
}

static void ReadProfilingBlock(const char *ToolName, FILE *F,
                               bool ShouldByteSwap,
                               std::vector<unsigned> &Data) {
  // Read the number of entries...
  unsigned NumEntries;
  if (fread(&NumEntries, sizeof(unsigned), 1, F) != 1) {
    errs() << ToolName << ": data packet truncated!\n";
    perror(0);
    exit(1);
  }
  NumEntries = ByteSwap(NumEntries, ShouldByteSwap);

  // Read the counts...
  std::vector<unsigned> TempSpace(NumEntries);

  // Read in the block of data...
  if (fread(&TempSpace[0], sizeof(unsigned)*NumEntries, 1, F) != 1) {
    errs() << ToolName << ": data packet truncated!\n";
    perror(0);
    exit(1);
  }

  //outs() << "TempSpace Print\n";
  /*
  for(unsigned int index = 0; index < TempSpace.size(); index++) {
//  	outs() << ToolName << "index: " << index << "\tTS: %u", TempSpace[index] << "\n";
  	printf("%s: index: %u, TS: %u\n", ToolName, index, TempSpace[index]);
  }
  */
  //outs() << "TempSpace Print: Done\n";

  // Make sure we have enough space... The space is initialised to -1 to
  // facitiltate the loading of missing values for OptimalEdgeProfiling.
  if (Data.size() < NumEntries)
    Data.resize(NumEntries, ProfileInfoLoader::Uncounted);

  // Accumulate the data we just read into the data.
  if (!ShouldByteSwap) {
    for (unsigned i = 0; i != NumEntries; ++i) {
      Data[i] = AddCounts(TempSpace[i], Data[i]);
    }
  } else {
    for (unsigned i = 0; i != NumEntries; ++i) {
      Data[i] = AddCounts(ByteSwap(TempSpace[i], true), Data[i]);
    }
  }
  //outs() << "Data Print\n";
  /*
  for(unsigned int index = 0; index < Data.size(); index++) {
  	printf("%s: index: %u, Data: %u\n", ToolName, index, Data[index]);
  }
  */
  //outs() << "Data Print: Done\n";
}


static void ReadProfilingBlock_AP(const char *ToolName, FILE *F,
				  bool ShouldByteSwap,
				  std::vector<unsigned> &Data) {

    unsigned NumEntries;
    if (fread(&NumEntries, sizeof(unsigned), 1, F) != 1) {
    	errs() << ToolName << ": data packet truncated!\n";
    	perror(0);
    	exit(1);
    }
    NumEntries = ByteSwap(NumEntries, ShouldByteSwap);

    // Read the counts...
    std::vector<unsigned> TempSpace(NumEntries);

    // Read in the block of data...
    // Hacked to figure out how many entries are exactly in the file.
/*    unsigned int tempOfTemp;
    unsigned int countingSize = 0;
    while(fread(&tempOfTemp, sizeof(unsigned), 1, F) != 1) {
	TempSpace[countingSize] = tempOfTemp;
	countingSize++;
    } */


    if (fread(&TempSpace[0], sizeof(unsigned)*NumEntries, 1, F) != 1) {
    	errs() << ToolName << ": data packet truncated!\n";
    	perror(0);
    	exit(1);
    } 

  // Make sure we have enough space... The space is initialised to -1 to
  // facitiltate the loading of missing values for OptimalEdgeProfiling.

/*  if(NumEntries != countingSize) {
	NumEntries = countingSize;
  } */
  if (Data.size() < NumEntries)
    	Data.resize(NumEntries, ProfileInfoLoader::Uncounted); 


  // Accumulate the data we just read into the data.
  if (!ShouldByteSwap) {
    for (unsigned i = 0; i != NumEntries; ++i) {
      Data[i] = TempSpace[i];
    }
  } else {
    for (unsigned i = 0; i != NumEntries; ++i) {
      Data[i] = ByteSwap(TempSpace[i], true);
    }
  }
}

static void ReadProfilingBlock_Array(const char *ToolName, FILE *F,
				  bool ShouldByteSwap,
				  std::vector<unsigned> &Data) {

    unsigned NumEntries;
    if (fread(&NumEntries, sizeof(unsigned), 1, F) != 1) {
    	errs() << ToolName << ": data packet truncated!\n";
    	perror(0);
    	exit(1);
    }
    NumEntries = ByteSwap(NumEntries, ShouldByteSwap);

    // Read the counts...
    std::vector<unsigned> TempSpace(NumEntries);

    if (fread(&TempSpace[0], sizeof(unsigned)*NumEntries, 1, F) != 1) {
    	errs() << ToolName << ": data packet truncated!\n";
    	perror(0);
    	exit(1);
    } 

  // Make sure we have enough space... The space is initialised to -1 to
  // facitiltate the loading of missing values for OptimalEdgeProfiling.

  if (Data.size() < NumEntries)
    	Data.resize(NumEntries, ProfileInfoLoader::Uncounted); 


  // Accumulate the data we just read into the data.
  if (!ShouldByteSwap) {
    for (unsigned i = 0; i != NumEntries; ++i) {
      Data[i] = TempSpace[i];
    }
  } else {
    for (unsigned i = 0; i != NumEntries; ++i) {
      Data[i] = ByteSwap(TempSpace[i], true);
    }
  }
}

static void ReadProfilingBlock_Float(const char *ToolName, FILE *F,
				  bool ShouldByteSwap,
				  IntFloatMap &Data) {

    unsigned NumEntries;
    if (fread(&NumEntries, sizeof(unsigned), 1, F) != 1) {
    	errs() << ToolName << ": data packet truncated!\n";
    	perror(0);
    	exit(1);
    }
    NumEntries = ByteSwap(NumEntries, ShouldByteSwap);
	
	unsigned int numRead = 0;

	while(numRead < NumEntries) {
		unsigned int PrimeID;
		float AliasProb;

		if(fread(&PrimeID, sizeof(unsigned), 1, F) != 1) {
			errs() << ToolName << ": data packet truncated!\n";
    		perror(0);
    		exit(1);
		}
		numRead++;

		if(fread(&AliasProb, sizeof(float), 1, F) != 1) {
			errs() << ToolName << ": data packet truncated!\n";
    		perror(0);
    		exit(1);
		}
		numRead++;
		
		outs() << "PrimeID: " << PrimeID << ", AliasProbability: " << AliasProb << "\n";

		Data[PrimeID] = AliasProb;
	}
}


static void ReadProfilingBlock_Yew(const char *ToolName, FILE *F,
				  bool ShouldByteSwap,
				  TwoLevelIntMap &Data) {

	unsigned int RefID;
	// First read in the RefID
	if (fread(&RefID, sizeof(unsigned), 1, F) != 1) {
   	 	errs() << ToolName << ": data packet truncated!\n";
   		perror(0);
    	exit(1);
    }
	do{

	    unsigned NumEntries;    
		if (fread(&NumEntries, sizeof(unsigned), 1, F) != 1) {
   		 	errs() << ToolName << ": data packet truncated!\n";
   			perror(0);
   	 		exit(1);
   	 	}

		NumEntries = ByteSwap(NumEntries, ShouldByteSwap);

		//outs() << "RefID: " << RefID << ", NumEntries: " << NumEntries << "\n";

		if(NumEntries == 0) {
			continue;
		}
		
		assert(!(NumEntries % 2) && "NumEntries is not a multiple of 2");
	    // Read the counts...
   	 	std::vector<unsigned> TempSpace(NumEntries);

		if (fread(&TempSpace[0], sizeof(unsigned)*NumEntries, 1, F) != 1) {
    		errs() << ToolName << ": data packet truncated!\n";
    		perror(0);
    		exit(1);
    	} 

		// The layout in the file would be:
		// YewDDPInfo | RefID | NumEntries | RefID-in-map-1 | RefID-in-map-1 count | Ref-in-map-2 | RefID-in-map-2 count 
		
		// Layout in TempSpace
		// 			0		  |   1   | 		2		   |   3
		// RefID-in-map-index | count | RefID-in-map-index | count

	  	// Accumulate the data we just read into the data.
	  	if (!ShouldByteSwap) {
   	  		for (unsigned i = 0; i != NumEntries; ++i) {
				if(i % 2) {
					//outs() << TempSpace[i-1] << ", " << TempSpace[i] << "\n";
      				Data[RefID][TempSpace[i-1]] = TempSpace[i];
				}
    		}
  	  	} else {
   	  		for (unsigned i = 0; i != NumEntries; ++i) {
				if(i % 2) {
					unsigned int RefID_in_map = ByteSwap(TempSpace[i-1], true);
					unsigned int count = ByteSwap(TempSpace[i], true);
					//outs() << RefID_in_map << ", " << count << "\n";
      				Data[RefID][RefID_in_map] = count;
				}
    		}
  	  	}

	}while(fread(&RefID, sizeof(unsigned), 1, F) == 1);
}



const unsigned ProfileInfoLoader::Uncounted = ~0U;

// ProfileInfoLoader ctor - Read the specified profiling data file, exiting the
// program if the file is invalid or broken.
//
ProfileInfoLoader::ProfileInfoLoader(const char *ToolName,
                                     const std::string &Filename,
                                     Module &TheModule) :
                                     Filename(Filename), 
                                     M(TheModule), Warned(false) {


  FILE *F = fopen(Filename.c_str(), "rb");

  if(F == 0) {
    errs() << ToolName << ": Error opening '" << Filename << "': ";
    perror(0);
    exit(1);
  }

  // Keep reading packets until we run out of them.
  unsigned PacketType;
  while (fread(&PacketType, sizeof(unsigned), 1, F) == 1) {
    // If the low eight bits of the packet are zero, we must be dealing with an
    // endianness mismatch.  Byteswap all words read from the profiling
    // information.
    bool ShouldByteSwap = (char)PacketType == 0;
    PacketType = ByteSwap(PacketType, ShouldByteSwap);

    switch (PacketType) {
    case ArgumentInfo: {
      unsigned ArgLength;
      if (fread(&ArgLength, sizeof(unsigned), 1, F) != 1) {
        errs() << ToolName << ": arguments packet truncated!\n";
        perror(0);
        exit(1);
      }
      ArgLength = ByteSwap(ArgLength, ShouldByteSwap);

      // Read in the arguments...
      std::vector<char> Chars(ArgLength+4);

      if (ArgLength)
        if (fread(&Chars[0], (ArgLength+3) & ~3, 1, F) != 1) {
          errs() << ToolName << ": arguments packet truncated!\n";
          perror(0);
          exit(1);
        }
      CommandLines.push_back(std::string(&Chars[0], &Chars[ArgLength]));
      break;
    }

    case FunctionInfo:
	  //outs() << "FunctionCounts -- Filling Up\n";
      ReadProfilingBlock(ToolName, F, ShouldByteSwap, FunctionCounts);
      break;

    case BlockInfo:
	  //outs() << "BlockCounts -- Filling Up\n";
	  BlockCounts.clear();
      ReadProfilingBlock(ToolName, F, ShouldByteSwap, BlockCounts);
      break;

    case EdgeInfo:
	  //outs() << "EdgeCounts -- Filling Up\n";
      ReadProfilingBlock(ToolName, F, ShouldByteSwap, EdgeCounts);
      break;

    case OptEdgeInfo:
      ReadProfilingBlock(ToolName, F, ShouldByteSwap, OptimalEdgeCounts);
      break;

    case BBTraceInfo:
      ReadProfilingBlock(ToolName, F, ShouldByteSwap, BBTrace);
      break;
    
    case AddrProfInfo:
      DEBUG(errs() << "AddrProf\n");
      ReadProfilingBlock_AP(ToolName, F, ShouldByteSwap, AddrProfCounts);
      DEBUG(errs() << "Done with AddrProf\n");
      break;

    case PreheaderInfo:
      DEBUG(errs() << "PreheaderInfo\n");
      ReadProfilingBlock_AP(ToolName, F, ShouldByteSwap, PreheaderCounts); 
      break;

    case LoadOccurenceInfo:
      ReadProfilingBlock_AP(ToolName, F, ShouldByteSwap, LoadOccurenceCounts);
      break;
	
	case YewDDPInfo:
	  //outs() << "YewInfo\n";
	  ReadProfilingBlock_Yew(ToolName, F, ShouldByteSwap, YewProfCounts);
	  break;

	case ArrayInfo:
	  //outs() << "ArrayInfo\n";
	  ReadProfilingBlock_Array(ToolName, F, ShouldByteSwap, ArrayProfCounts);
	  break;
    
	case PinInfo:
	  outs() << "PinInfo\n";
	  ReadProfilingBlock_Float(ToolName, F, ShouldByteSwap, PinCounts);
	  break;

    default:
	  printf("Unknown packet header\n");
      errs() << ToolName << ": Unknown packet type #" << PacketType << "!\n";
      exit(1);
    }
  }

  fclose(F);
}

