//===-- FunctionCloning.cpp - Clone functions for Spec - Opts -------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "funcclone"
#include "llvm/Transforms/IPO.h"
#include "llvm/Constant.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Support/CallSite.h"
#include "llvm/Support/Compiler.h"
#include "llvm/ADT/DenseSet.h"
#include "llvm/Analysis/Dominators.h"
//#include "llvm/Analysis/SCCTraversal.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Analysis/AliasAnalysis.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/ProfileInfoLoader.h"
#include "llvm/Support/CommandLine.h"
//#include "llvm/Transforms/Utils/ValueMapper.h"
#include <map>
using namespace llvm;


// Naive scheme. First five functions are selected
#define SELECT_MAX 1
#define ENABLE_SELECT_MAX
#define ENABLE_SKIP_MAIN
#define FUNCTION_FILTER
#define SPEC_THRESH 0.5
//#define ADD_TO_SPECLIST
//#define VOID_ONLY

static cl::opt<std::string>
FunctionName("fn-name", cl::init("main"),
                    cl::value_desc("filename"),
                    cl::desc("Function to be cloned using FuncClone"));

static cl::opt<bool>
ProfileRun("profrun", cl::Hidden,
			 cl::desc("Profile Run for finding suitable pairs in Function Cloning"), cl::init(false));


namespace {
  class FuncClone : public ModulePass {
	

	typedef std::vector<Function*> FunctionStack;
	FunctionStack SelectedFunctions;
	Module *CurrModule;
	LLVMContext *CurrContext;
	DominatorTree *DT;
	DominanceFrontier *DF;
	AliasAnalysis *AA;
	unsigned int PrimeIDCount;
	//ProfileInfoLoader PinPIL;
	//AliasSetTracker *CurAST;

  public :
    static char ID; // Pass identification, replacement for typeid
    FuncClone() : ModulePass(&ID){} //, PinPIL("Profiling Feedback to FuncClone", "ProfFB.out", *CurrModule) {}
	virtual void getAnalysisUsage(AnalysisUsage &AU) const {
	  AU.addRequired<DominatorTree>();
	  AU.addRequired<DominanceFrontier>();
	  //AU.addRequired<SCCTraversal>();
	  AU.addRequired<AliasAnalysis>();
    }

    bool runOnModule(Module &M);
	void FuncSelect();
	void CloneAndInsert();
	virtual const char *getPassName() const { return "Function Cloning"; }
  };
}

char FuncClone::ID = 0;
static RegisterPass<FuncClone> WD("funcclone", "Function Cloning");

bool FuncClone::runOnModule(Module &M) {

  bool Changed = false;

  PrimeIDCount = 1;
  
  outs() << "FuncClone: Begin\n";
  AA = &getAnalysis<AliasAnalysis>();

  CurrModule = &M;

  // Function Selection
  FuncSelect();

  // Function Cloning
  // Making changes at call sites
  CloneAndInsert();
  
  outs() << "FuncClone: End\n"; 

  return true;
}


// Strategy for selecting functions
void FuncClone::FuncSelect() {

	outs() << "Selected Functions: Begin\n";
	unsigned int count = 0;
	#ifdef ENABLE_SKIP_MAIN
	Function *Main = CurrModule->getFunction("main");
	if(Main->getName() != "main") {
		outs() << "No main function!\n";
	}
	#endif

	for(Module::iterator mit = CurrModule->begin(), mitend = CurrModule->end(); mit != mitend; ++mit) {
		
		// If its just a declaration of the function or if its already been cloned, or if its a clone of another function...
		// nothing to do
		if(mit->isDeclaration() || mit->isCloned() || mit->isItAClone()) {
			continue;
		}

		#ifdef FUNCTION_FILTER
		if(mit->getName() != FunctionName) {
			continue;
		}
		#if 0
		// Gcc
		if(mit->getName() != "propagate_block") {
			continue;
		}
		#endif
		#if 0
		// Gap
		if(mit->getName() != "EvAssList") {
			continue;
		}
		#endif
		#if 0
		// Parser
		if(mit->getName() != "prune_match") {
			continue;
		}
		#endif
		#if 0
		// Twolf
		if(mit->getName() != "new_dbox_a") {
			continue;
		}
		#endif
		#if 0
		// Vpr 
		if(mit->getName() != "try_swap") {
			continue;
		}
		#endif
		#if 0
		// Gzip 
		if(mit->getName() != "inflate_codes") {
			errs() << "mit: " << mit->getName() << "\n";
			continue;
		}
		else {
			errs() << "Found deflate\n";
		}
		#endif
		#if 0
		// Bzip2 
		if(mit->getName() != "sortIt") {
			continue;
		}
		#endif
		#if 0
		// Mesa 
		if(mit->getName() != "gl_depth_test_span_less") {
			continue;
		}
		#endif
		#if 0
		// Ammp 
		if(mit->getName() != "eval") {
			continue;
		}
		#endif	
		#if 0
		// Art 
		if(mit->getName() != "train_match") {
			continue;
		}
		#endif
		#if 0
		// Swim 
		if(mit->getName() != "calc1_") {
			continue;
		}
		#endif

		#endif

		#ifdef ENABLE_SKIP_MAIN
		if(mit->getName() == "main") {
			outs() << "Skipping main\n";
			continue;
		}
		#endif

		// If the function is not called through a CallInst each time, we don't consider it.
		bool UsesOkay = true;
		for(Value::use_iterator uit = mit->use_begin(), uitend = mit->use_end(); uit != uitend; ++uit) {
			Instruction* CurrInst = dyn_cast<Instruction>(*uit);
			if(!CurrInst) {
				UsesOkay = false;
				break;
			}
			if(!isa<CallInst>(CurrInst) && !isa<InvokeInst>(CurrInst)) {
				UsesOkay = false;
				break;
			}

			#ifdef VOID_ONLY
			CurrContext = &mit->getContext();
			if(CurrInst->getType() != Type::getVoidTy(*CurrContext)) {
				UsesOkay = false;
				break;
			}
			#endif

			CallSite CS(CurrInst);	
			if(CS.getCalledFunction() != mit) {
				UsesOkay = false;
				break;
			}
		}
		if(!UsesOkay) {
			#ifdef FUNCTION_FILTER
			outs() << "Function Not Selected since all uses are not CallInsts\n";
			exit(0);
			#endif
			continue;
		}


		outs() << mit->getName() << "\n";
		SelectedFunctions.push_back(mit);
		count++;
		#ifdef ENABLE_SELECT_MAX
		if(count >= SELECT_MAX) {
			#if 0
			#ifdef FUNCTION_FILTER
			outs() << "Select Max Enabled. Exceeded max # of fns.\n";
			exit(0);
			#endif
			#endif
			break;
		}
		#endif
	}
	outs() << "Selected Functions: End\n";
}

// RemapInstruction - Convert the instruction operands from referencing the
// current values into those specified by ValueMap. - LIFTED RIGHT OUT OF LoopUnswitch.cpp where it
// is also declared as a static function -- Works for a single basic block only!!!
//
static inline void RemapInstruction(Instruction *I,
                                    DenseMap<const Value *, Value*> &ValueMap) {
  for (unsigned op = 0, E = I->getNumOperands(); op != E; ++op) {
    Value *Op = I->getOperand(op);
    DenseMap<const Value *, Value*>::iterator It = ValueMap.find(Op);
    if (It != ValueMap.end()) Op = It->second;
    I->setOperand(op, Op);
  }
}

// Clone the function, and insert necessary code at all call sites
void FuncClone::CloneAndInsert() {


	// Convention
	// CurrFn	Spec
	// NewFn	Nonspec
#if 0
	const std::string Filename = "ProfFB.out";

	FILE *F = fopen(Filename.c_str(), "rb");

  	if(F == 0) {
    	errs() << ToolName << ": Error opening '" << Filename << "': ";
  	}
	else {
#endif

	ProfileInfoLoader PinPIL("Profiling Feedback to FuncClone", "Prof.FB_out", *CurrModule);
	if(ProfileRun) {
		outs() << "FunctionCloning: THIS IS A PROFILE RUN\n";
	}
	else {
		outs() << "FunctionCloning: THIS IS A SPECULATION RUN WITH SPEC_THRESH: " << SPEC_THRESH << "\n";
	}
	
#ifdef ADD_TO_SPECLIST
	outs() << "FunctionCloning: ADD_TO_SPECLIST Enabled\n";
#endif

	// Iterate over all the functions selected
	for(FunctionStack::iterator mit = SelectedFunctions.begin(), mitend = SelectedFunctions.end(); mit != mitend; ++mit) {

		Function* CurrFn = *mit;
		CurrModule = CurrFn->getParent();
		CurrContext = &CurrFn->getContext();

		// Now, clone the function
		// Passing an empty DenseMap should cause CloneFunction to return an identical clone to the current function
		DenseMap<const Value*, Value*> DummyMap;
		DummyMap.clear();
		Function *NewFn = CloneFunction(CurrFn, DummyMap);
		NewFn->setLinkage(CurrFn->getLinkage());
  		CurrFn->getParent()->getFunctionList().push_back(NewFn);
		
		CurrFn->setCloned();
		NewFn->setItsAClone(CurrFn);

		outs() << "CurrFn: " << CurrFn->getName() << "\tNewFn: " << NewFn->getName() << "\n";

		// Insert instructions at all the call sites of the function -- 
		// Also need to insert instructions on all blocks within the cloned function containing a retInst for 
		// checking speculation.


		for(Value::use_iterator uit = CurrFn->use_begin(), uitend = CurrFn->use_end(); uit != uitend; ++uit) {
			

			// At call site
			// spec = ShouldSpeculate();
			// if(spec) {
			// 		jump to cloned function;
			// }
			// else {
			// 		jump to original function;
			// }
		
			Instruction* CurrInst = cast<Instruction>(*uit);

			BasicBlock* CurrBB = CurrInst->getParent();


			// Lets now split the current basic block at the CallInst. The BB thus created will be called NonSpecBB.
//			BasicBlock *NonSpecBB = SplitBlock(CurrBB, CurrInst, this);
			BasicBlock *SpecBB = SplitBlock(CurrBB, CurrInst, this);

			// Now, split the NonSpecBB immediately at the second instruction
//			BasicBlock::iterator bbit = NonSpecBB->begin();
			BasicBlock::iterator bbit = SpecBB->begin();
			bbit++; // This now points to the second inst in the basicblock
			// Now that we know the second inst, split it at that inst. The new block will be called NewBB
//			BasicBlock* NewBB = SplitBlock(NonSpecBB, bbit, this);
			BasicBlock* NewBB = SplitBlock(SpecBB, bbit, this);

			// We should now have three blocks in place of the one block before. The first (CurrBB) unconditionally jumps
			// to the second block (SpecBB), and the second unconditionally jumps to the third block (NewBB).
			


			// Now we clone the SpecBB to create the NonSpecBB. This should create a basic block with an identical call inst
			// and jump inst. 
			DenseMap<const Value*, Value*> LocalDummyMap;
//			BasicBlock *SpecBB = CloneBasicBlock(NonSpecBB, LocalDummyMap, "Spec", NonSpecBB->getParent());
			BasicBlock *NonSpecBB = CloneBasicBlock(SpecBB, LocalDummyMap, "NonSpec", SpecBB->getParent());

			// Then, we change the function call in the SpecBB to point to the NewFn.
			// CODE TAKEN FROM SpecializeFunction
			Instruction* SpecBBFirstInst = SpecBB->begin();
			CallSite SpecCS(cast<Instruction>(SpecBBFirstInst));	
			assert(SpecCS.getCalledFunction() == CurrFn);
			CallSite NonSpecCS(cast<Instruction>(NonSpecBB->begin()));
			// No PHINodes have crept in. Just a safety check 
			assert(SpecBBFirstInst == SpecBB->getFirstNonPHI());			
			Value *NewFnCall;
			SmallVector<Value*, 8> NewFnArgs;
			assert(SpecCS.arg_size() == NonSpecCS.arg_size());
			for(unsigned int index = 0, end = NonSpecCS.arg_size(); index < end; index++) {
				NewFnArgs.push_back(NonSpecCS.getArgument(index));
				assert(SpecCS.getArgument(index) == NonSpecCS.getArgument(index));
			}


			// Check if both the argument lists are the same?
			#if 0
			CallSite SpecCS(cast<Instruction>(SpecBBFirstInst));
			#endif


			//NewFn->copyAttributesFrom(CurrFn);
			NewFn->setAttributes(CurrFn->getAttributes());
			Instruction* NonSpecBBFirstInst = NonSpecBB->begin();
			if(CallInst *CI = dyn_cast<CallInst>(NonSpecBBFirstInst)) {
				NewFnCall = CallInst::Create(NewFn, NewFnArgs.begin(), NewFnArgs.end(), CI->getName(), CI);
				// ADDED 
				cast<CallInst>(NewFnCall)->setTailCall(CI->isTailCall());
				cast<CallInst>(NewFnCall)->setCallingConv(CI->getCallingConv());
			}
			else {
				assert(isa<InvokeInst>(NonSpecBBFirstInst));
				InvokeInst *II = cast<InvokeInst>(NonSpecBBFirstInst); 
				NewFnCall = InvokeInst::Create(NewFn, II->getNormalDest(), II->getUnwindDest(), NewFnArgs.begin(), NewFnArgs.end(), II->getName(), II);
				// ADDED
				cast<InvokeInst>(NewFnCall)->setCallingConv(II->getCallingConv());
			}
			// ReplaceInstWithInst(ClonedBBFirstInst, cast<Instruction>(NewFnCall)); -- This may not replace all uses properly
			NonSpecCS.getInstruction()->replaceAllUsesWith(NewFnCall);
			NonSpecCS.getInstruction()->eraseFromParent();

			// Lets check the uses for each of the arguments. The current SpecFunction and the NonSpecFn have to be listed
			for(unsigned int index = 0, end = NonSpecCS.arg_size(); index < end; index++) {
				bool NonSpecCSFound = false;
				// In the uses of the arguments of the SpecCS (the site of the original function, CurrFn), the NonSpec function
				// call should also be listed.
				for(Value::use_iterator local_uit = SpecCS.getArgument(index)->use_begin(), 
										local_uend = SpecCS.getArgument(index)->use_end(); local_uit != local_uend; local_uit++) {
						outs() << "Arg:" << index << "\tValue:" << **local_uit << "\n";
						if(Instruction* UseInst = dyn_cast<Instruction>(*local_uit)) {
							if(UseInst == NonSpecBB->begin()) {
								NonSpecCSFound = true;
								break;
							}
						}
				}
				if(!NonSpecCSFound) {
					assert("NonSpecCS not found in the users list\n");
				}
			}



			// Then, we insert a PHINode in NewBB to choose between the values from the CallInst in SpecBB and NonSpecBB
			// Then, replaceAllUses NonSpecBB's value with the new PHINode value.
			assert(SpecBB->begin()->getType() == NonSpecBB->begin()->getType() && "FC: SpecBB Type != NonSpecBB Type");
			// If the function returns a void type, we don't have to do anything.
			if(NonSpecBB->begin()->getType() != Type::getVoidTy(*CurrContext)) {	
				#ifdef VOID_ONLY
				assert(0);
				#endif
				PHINode* PN = PHINode::Create(NonSpecBB->begin()->getType(), "", NewBB->getFirstNonPHI());
				// This has to be done before adding the NonSpec to PHINode. Otherwise the use of NonSpec in the PHINode itself
				// is replaced!!!
				SpecBB->begin()->replaceAllUsesWith(PN);
				// Adding the Spec Function call to the PHINode
				Value *SpecV = cast<Value>(SpecBB->begin());
				assert(SpecV);
				PN->addIncoming(SpecV, SpecBB);
				// Adding the NonSpec Function call to the PHINode
				Value *NonSpecV = cast<Value>(NonSpecBB->begin());
				assert(NonSpecV);
				PN->addIncoming(NonSpecV, NonSpecBB);
			}
			// Then we change the Terminator in the CurrBB to become a branch between the two BBs
			Value* ConstantZero = ConstantInt::get(Type::getInt32Ty(*CurrContext), 0, false);
			Constant *PinRef = CurrModule->getOrInsertFunction("PinRefFn", Type::getInt32Ty(*CurrContext),(char*)NULL); //,
			//											Type::getInt32Ty(*CurrContext),
			//											(Type *)0);

			//std::vector<Value *>PinRefArgs(1);
			//PinRefArgs[0] = ConstantZero;
			//Instruction *PinRefCall = CallInst::Create(PinRef, PinRefArgs.begin(), PinRefArgs.end(), "PinRefRetVal", CurrBB->getTerminator());	
			Instruction *PinRefCall = CallInst::Create(PinRef, "PinRefRetVal", CurrBB->getTerminator());
		

			// The ICmp Inst compares the output value of the PinRefFn (PinRefCall) with zero. This will be used
			// in the branch inst coming up
			Value* BranchOutputVar = new ICmpInst(CurrBB->getTerminator(), ICmpInst::ICMP_EQ, PinRefCall, ConstantZero, "spec_opt_icmp");


			// PinRefCall = 1 => Branch Output = FALSE => Go to Spec => Branch will go to SpecBB => CurrFn 
			// PinRefCall = 0 => Branch Output = TRUE => Go to NonSpec => Branch will go to NonSpecBB => NewFn 
			BranchInst *BI = BranchInst::Create(NonSpecBB, SpecBB, BranchOutputVar, CurrBB->getTerminator());

			// Insert the branch instruction in place of the unconditional jump
//			ReplaceInstWithInst(CurrBB->getTerminator(), BI);
			CurrBB->getTerminator()->replaceAllUsesWith(BI);
			CurrBB->getTerminator()->eraseFromParent();


			// Insert MemChkPt before the PinRefFn call
			const Type* VoidTy   = Type::getVoidTy(*CurrContext);
			const Type* IntTy = Type::getInt32Ty(*CurrContext);
			Constant *PinMemChkPtFn = CurrModule->getOrInsertFunction("PinMemChkPt", VoidTy, (char*)NULL);
			Instruction* PinMemChkPt = CallInst::Create(PinMemChkPtFn, "", PinRefCall);

		
			// Insert the SpecCheck before the terminator of SpecBB. The other instruction right now is the call to the
			// cloned version of the original function
			#if 0
			Constant *PinSpecChkFn = CurrModule->getOrInsertFunction("PinSpecCheck", VoidTy, (char*)NULL);
			Instruction* PinSpecChk = CallInst::Create(PinSpecChkFn, "", SpecBB->getTerminator());
			#endif


			BasicBlock &EntryBlock = (CurrBB->getParent())->getEntryBlock();
			// Insert the branch below only if the block calling the function is not the EntryBB
			if((&EntryBlock) != CurrBB) {
					
				// We need to ensure that the PinSpecCheck is linked to the branch instruction --- 
				// Ideally we would like to hold onto something from the function call as well but we don't know if the function has 
				// any arguments/return values. What we are doing here is to put the create a IntTy return value for PinSpecCheck
				// which will be used to choose the branch.
				Constant *PinSpecChkFn = CurrModule->getOrInsertFunction("PinSpecCheck", IntTy, (char*)NULL);
				Instruction *PinSpecChk = CallInst::Create(PinSpecChkFn, "SpecCheckRetVal", SpecBB->getTerminator());
	
				Value *SpecChkBranchOutput = new ICmpInst(SpecBB->getTerminator(), ICmpInst::ICMP_EQ, PinSpecChk, ConstantZero, "SpecChk_icmp");
	
				// PinSpecChk = 1 => Branch Output = FALSE => Go to NewBB
				// PinSpecChk = 0 => Branch Output = TRUE  => Go to CurrBB
				BranchInst* SpecChkBI = BranchInst::Create(CurrBB, NewBB, SpecChkBranchOutput, SpecBB->getTerminator());
	
				SpecBB->getTerminator()->replaceAllUsesWith(SpecChkBI);
				SpecBB->getTerminator()->eraseFromParent();
	
				enum CallingConv::ID SpecChkCallConv = CallingConv::C;
				cast<CallInst>(PinSpecChk)->setCallingConv(SpecChkCallConv);
	
				// The CurrBB has PhiNodes which need to be updated to include the value from this path
				// Since we don't create any new values in the SpecBB, we simply put the value of each PhiNode as
				// the value copied over --- this should not create a use before def problem since PhiNode should be
				// able to handle the case where there is a loop, and you come through the header, and copy the value back
				for(BasicBlock::iterator pnit = CurrBB->begin(); pnit != CurrBB->end(); pnit++) {
					if(PHINode* CurrPN = dyn_cast<PHINode>(pnit)) {
						CurrPN->addIncoming(CurrPN, SpecBB);
					}
				}
			}
		}
		
		// Not using SCCTraversal. Instead, using the Alias Queries
#if 0
		// Use the SCCTraversal data for the CurrFn, and speculate on a pair of load and store
		for(int index = CurrFn->SCCTLoads.find_first(); index != -1; index = CurrFn->SCCTLoads.find_next(index)) {

			LoadInst* LI = cast<LoadInst>(CurrFn->ReverseMapping[index]);

			// Put this load in set 1
			const Type* VoidTy   = Type::getVoidTy(*CurrContext);
			const Type* LoadTy = LI->getOperand(0)->getType();
			const Type* IntTy = Type::getInt32Ty(*CurrContext);	
			Constant *Pin_Set_InsertFn = CurrModule->getOrInsertFunction("PinSetInsert", VoidTy, 
																	LoadTy,
																    IntTy,	
																	(char*)NULL);

			std::vector<Value*> Args(2);
			Args[0] = LI->getOperand(0);
			Args[1] = ConstantInt::get(Type::getInt32Ty(*CurrContext), 1, false);
			Instruction *SetInsertInst = CallInst::Create(Pin_Set_InsertFn, Args.begin(), Args.end(),"", LI);


			for(DenseMap<unsigned int, LoadInst::StoreSet>::iterator dit = LI->AliasSets.begin();
				dit != LI->AliasSets.end(); ++dit) {
				
				BitVector AliasBV = dit->second;

				for(int index2 = AliasBV.find_first(); index2 != -1; index2 = AliasBV.find_next(index2)) {
					
					StoreInst *SI = cast<StoreInst>(CurrFn->ReverseMapping[index2]);

					// Put this store in set 2
					const Type* StoreTy = SI->getOperand(1)->getType();
					Pin_Set_InsertFn = CurrModule->getOrInsertFunction("PinSetInsert", VoidTy, 
																	StoreTy,
																	IntTy,
																	(char*)NULL);

					Args[0] = SI->getOperand(1);
					Args[1] = ConstantInt::get(Type::getInt32Ty(*CurrContext), 2, false);
					SetInsertInst = CallInst::Create(Pin_Set_InsertFn, Args.begin(), Args.end(),"", SI);

					// Add each load store pair to SpecNoAlias list
					unsigned int LISize = 0;
					LISize = AA->getTypeStoreSize(LI->getType());

					unsigned int SISize = 0;
					SISize = AA->getTypeStoreSize(SI->getOperand(0)->getType());

					AA->addToSpecNoAliases(LI->getOperand(0), SI->getOperand(1), LISize, SISize);
					outs() << "LD: " << *LI << "\nST: " << *SI << "\n";
				}

			}

			// Stopping after the first load
			//break;

		}
#endif

	
#if 0
		unsigned int count = 0;
		//for(unsigned int index = 0; index < AA->Queries.size(); index++) {
		for(AliasAnalysis::ASMap::iterator asit = AA->Queries.begin(); asit != AA->Queries.end(); asit++) {
			const Value *V1 = asit->first;
			AliasAnalysis::ASVec *CurrASVec = &(asit->second);
			for(unsigned int index = 0; index < CurrASVec->size(); index++) {

			//AliasStruct* CurrAS = AA->Queries[index];
				AliasStruct *CurrAS = (*CurrASVec)[index];
			/*
			if((Instruction* I1 = dyn_cast<Instruction>(CurrAS->V1)) && 
				(Instruction* I2 = dyn_cast<Instruction>(CurrAS->V2))) {
			*/
				if(const Instruction* I1 = dyn_cast<Instruction>(CurrAS->V1)) {
					if(const Instruction* I2 = dyn_cast<Instruction>(CurrAS->V2)) {
						const BasicBlock* BB1 = I1->getParent();
						const BasicBlock* BB2 = I2->getParent();
						const Function* F1 = BB1->getParent();
						const Function* F2 = BB2->getParent();
						if((F1 == CurrFn) && (F2 == CurrFn)) {
							AA->addToSpecNoAliases(CurrAS->V1, CurrAS->V2, CurrAS->V1Size, CurrAS->V2Size);
							outs() << "SpecNoAliases\n";
							outs() << "V1: " << *(CurrAS->V1) << "\nV2: " << *(CurrAS->V2) << "\n";
							count++;

							// Instrument each of the instructions in the pair -- to a unique PrimeID
							unsigned int CurrPrimeID = PrimeIDCount++;
							
							// The V1 goes to the Prime, V2 to the Subprime.

							// Get the prototypes and args setup
							// V1
							const Type* VoidTy   = Type::getVoidTy(*CurrContext);
							const Type* IntTy = Type::getInt32Ty(*CurrContext);	
							Constant *PinPrimeInsertFn = CurrModule->getOrInsertFunction("PinPrimeInsert", VoidTy, 
																	I1->getType(), // Address
																	IntTy, // CurrPrimeID
																	(char*)NULL);
							
							
							std::vector<Value*> PrimeArgs(2);

							// Find the non-constant version of V1 & V2
							Value *nonConstV1;
							Value *nonConstV2;
							Value *nonConstI1;
							Value *nonConstI2;
							for(Function::iterator fit = CurrFn->begin(); fit != CurrFn->end(); fit++) {
								if(fit == BB1) {
									for(BasicBlock::iterator bbit = fit->begin(); bbit != fit->end(); bbit++) {
										if(bbit == I1) {
											nonConstI1 = bbit;
											nonConstV1 = cast<Value>(bbit);
										}
									}
								}
								if(fit == BB2) {
									for(BasicBlock::iterator bbit = fit->begin(); bbit != fit->end(); bbit++) {
										if(bbit == I2) {
											nonConstI2 = bbit;
											nonConstV2 = cast<Value>(bbit);
										}
									}
								}
							}
											

							//PrimeArgs[0] = CurrAS->V1;
							PrimeArgs[0] = nonConstV1;
							PrimeArgs[1] = ConstantInt::get(Type::getInt32Ty(*CurrContext), CurrPrimeID, false);
							
							// Instrument for V1
							outs() << "nonConstV1: " << *nonConstV1 << ", Uses: " << nonConstV1->getNumUses() << "\n";
							for(Value::use_iterator uit = nonConstI1->use_begin(), uitend = nonConstI1->use_end(); uit != uitend; ++uit) {
		
								Instruction* V1UseInst = cast<Instruction>(*uit);
								if(CallInst* CI = dyn_cast<CallInst>(V1UseInst)) {
									//outs() << "CI->getOperand(0): " << *(CI->getOperand(0)) << "\n";
									if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinSubprimeInsert") != 0) {
										continue;
									}
									if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinPrimeInsert") != 0) {
										continue;
									}
								}

								outs() << "Use: " << *V1UseInst << "\n";
								
								Instruction *PinPrimeInsertInst = 
									CallInst::Create(PinPrimeInsertFn, PrimeArgs.begin(), PrimeArgs.end(),"", V1UseInst);
							}
							outs() << "-----------------------------------------------------------------\n";
							// Prototypes and Args - V2
							Constant *PinSubprimeInsertFn = CurrModule->getOrInsertFunction("PinSubprimeInsert", VoidTy, 
																	I2->getType(), // Address
																	IntTy, // CurrPrimeID
																	(char*)NULL);
							
							
							std::vector<Value*> SubprimeArgs(2);
							//SubprimeArgs[0] = CurrAS->V2;
							SubprimeArgs[0] = nonConstV2;
							SubprimeArgs[1] = ConstantInt::get(Type::getInt32Ty(*CurrContext), CurrPrimeID, false);
							
							// Instrument for V2
							outs() << "nonConstV2: " << *nonConstV2 << ", Uses: " << nonConstV2->getNumUses() << "\n";
							for(Value::use_iterator uit = nonConstI2->use_begin(), uitend = nonConstI2->use_end(); uit != uitend; ++uit) {
		
								Instruction* V2UseInst = cast<Instruction>(*uit);
								if(CallInst* CI = dyn_cast<CallInst>(V2UseInst)) {
									//outs() << "CI->getOperand(0): " << *(CI->getOperand(0)) << "\n";
									if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinSubprimeInsert") != 0) {
										continue;
									}
									if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinPrimeInsert") != 0) {
										continue;
									}
								}
								outs() << "Use: " << *V2UseInst << "\n";
								
								Instruction *PinSubprimeInsertInst = 
									CallInst::Create(PinSubprimeInsertFn, SubprimeArgs.begin(), SubprimeArgs.end(),"", V2UseInst);
							}
							outs() << "-----------------------------------------------------------------\n";

						}
					}
					else {
						outs() << "I2 not an inst: " << *(CurrAS->V2) << "\n";
					}
				}
				else {
					outs() << "I1 not an inst: " << *(CurrAS->V1) << "\n";
				}
			}
		}
		outs() << "SpecNoAliases-size(): " << count << "\n";
#endif
#if 1
		//TODO: Set the CurrFn using setAAFunction before adding to SpecNoAliases.
		//********************************
		//TODO: Need to assign CurrPrimeID for each V1-V2 pair instead of assigning it to
		// one V1, and all V2's associated with it share the PrimeID. Thats how it currently
		// setup.
		//*******************************
		AA->setFunction(CurrFn);
		AliasAnalysis::FnASMap::iterator fait = AA->Queries.find(CurrFn);
		unsigned int count = 0;
		if(fait != AA->Queries.end()) {
			AliasAnalysis::ASMap::iterator asit;
			for(asit = AA->Queries[CurrFn]->begin(); asit != AA->Queries[CurrFn]->end(); asit++) {
				const Value* V1 = asit->first;
				#if 0
				outs() << "V1: " << *V1 << "\n";
				#endif
				AliasAnalysis::ASVec* CurrASVec = asit->second;
				unsigned int CurrPrimeID = PrimeIDCount++;
				IntFloatMap::iterator ifit = PinPIL.PinCounts.find(CurrPrimeID);
				if(ifit == PinPIL.PinCounts.end()) {
					#if 0
					outs() << "No entry for CurrPrimeID:" << CurrPrimeID << "\n";
					#endif
					//TODO: On pairs which we don't know anything about, we should not speculate
					// These could be pairs which are always misspeculated in the profiling run
					// and thats the reason they didnt' show up in the PinCounts. Until the PinCoutns
					// is changed to include *all* pairs, we should not speculate on pairs for which we
					// can't find entries in the PinCounts.
				}
				else {
					outs() << "Found CurrPrimeID: " << CurrPrimeID << ", AliasProb: " << PinPIL.PinCounts[CurrPrimeID] << "\n";
					// If this is not a profile run, then do not speculate on this V1, if its probability is greater
					// than the SPECULATION_THRESHOLD
					if(!ProfileRun && (PinPIL.PinCounts[CurrPrimeID] > SPEC_THRESH)) {
						outs() << "!Profile Run. PinPIL[" << CurrPrimeID << "] > " << SPEC_THRESH << "\n";
						continue;
					}
						
				}
				// Get the prototypes and args setup
				// V1
				const Type* VoidTy   = Type::getVoidTy(*CurrContext);
				const Type* IntTy = Type::getInt32Ty(*CurrContext);	
				Constant *PinPrimeInsertFn = CurrModule->getOrInsertFunction("PinPrimeInsert", VoidTy, 
															V1->getType(), // Address
															IntTy, // CurrPrimeID
															(char*)NULL);
							
							
				std::vector<Value*> PrimeArgs(2);

				// Here, we instrument the uses of V1 within this function
				for(Value::use_const_iterator uit = V1->use_begin(), uitend = V1->use_end(); uit != uitend; ++uit) {
					if(!(isa<Instruction>(*uit))) {
						continue;
					}
					const Instruction* V1UseInst = cast<Instruction>(*uit);
					if(isa<PHINode>(V1UseInst)) {
						#if 0
						outs() << "PHINode(V1): " << *V1UseInst << "\n";
						#endif
						continue;
					}
					if(const CallInst* CI = dyn_cast<CallInst>(V1UseInst)) {
						//outs() << "CI->getOperand(0): " << *(CI->getOperand(0)) << "\n";
						if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinSubprimeInsert") != 0) {
							continue;
						}
						if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinPrimeInsert") != 0) {
							continue;
						}
					}

					if(V1UseInst->getParent()->getParent() == CurrFn) {
						// This use of V1 belongs to this function -- Find the nonConst version of it, and instrument it
						Instruction *nonConstI1;
						Value *nonConstV1;
						for(Function::iterator fit = CurrFn->begin(); fit != CurrFn->end(); fit++) {
							if(fit == V1UseInst->getParent()) {
								for(BasicBlock::iterator bbit = fit->begin(); bbit != fit->end(); bbit++) {
									if(bbit == V1UseInst) {
										nonConstI1 = bbit;
										// Find nonConstV1
										#if 0
										if(LoadInst* LI = dyn_cast<LoadInst>(bbit)) {
											assert(V1 == LI->getOperand(0));
											nonConstV1 = bbit->getOperand(0);
										}
										else if(StoreInst* SI = dyn_cast<StoreInst>(bbit)) {
											assert(V1 == SI->getOperand(1));
											nonConstV1 = SI->getOperand(1);
										}
										else {
										#endif
											// Iterate over the operands of the instruction to find V1
											for(unsigned int op_index = 0; op_index < nonConstI1->getNumOperands(); 
												op_index++) {
												if(nonConstI1->getOperand(op_index) == V1) {
													nonConstV1 = nonConstI1->getOperand(op_index);
													break;
												}
											}
										#if 0
										}
										#endif
									}
								}
							}
						}
						assert(nonConstI1 != NULL);
						assert(nonConstV1 != NULL);

						//Instrument this nonConstV1
						PrimeArgs[0] = nonConstV1;
						PrimeArgs[1] = ConstantInt::get(Type::getInt32Ty(*CurrContext), CurrPrimeID, false);
				 
						Instruction *PinPrimeInsertInst =
							CallInst::Create(PinPrimeInsertFn, PrimeArgs.begin(), PrimeArgs.end(),"", nonConstI1);
					}
				}


				for(AliasAnalysis::ASVec::iterator avit = CurrASVec->begin(); avit != CurrASVec->end(); ++avit) {
					AliasStruct* AS = *avit;
					const Value *V2 = AS->V2;

#ifdef ADD_TO_SPECLIST
					count++;
					AA->addToSpecNoAliases(AS->V1, AS->V2, AS->V1Size, AS->V2Size);
#endif
					#if 0
					outs() << "SpecNoAliases\n";
					outs() << "V1: " << *(AS->V1) << "\nV2: " << *(AS->V2) << "\n";
					#endif

					#if 0
					outs() << "V2: " << *V2 << "\n";
					#endif

					// Get the prototypes and args setup
					// V1
					const Type* VoidTy   = Type::getVoidTy(*CurrContext);
					const Type* IntTy = Type::getInt32Ty(*CurrContext);	
					Constant *PinSubprimeInsertFn = CurrModule->getOrInsertFunction("PinSubprimeInsert", VoidTy, 
																V2->getType(), // Address
																IntTy, // CurrPrimeID
																(char*)NULL);
								
								
					std::vector<Value*> SubprimeArgs(2);
	
					// Here, we instrument the uses of V1 within this function
					for(Value::use_const_iterator uit = V2->use_begin(), uitend = V2->use_end(); uit != uitend; ++uit) {
						if(!(isa<Instruction>(*uit))) {
							continue;
						}
						const Instruction* V2UseInst = cast<Instruction>(*uit);
						if(isa<PHINode>(V2UseInst)) {
							#if 0
							outs() << "PHINode(V2): " << *V2UseInst << "\n";
							#endif
							continue;
						}
						if(const CallInst* CI = dyn_cast<CallInst>(V2UseInst)) {
							//outs() << "CI->getOperand(0): " << *(CI->getOperand(0)) << "\n";
							if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinSubprimeInsert") != 0) {
								continue;
							}
							if(strcmp(CI->getOperand(0)->getNameStr().c_str(), "PinPrimeInsert") != 0) {
								continue;
							}
						}

						if(V2UseInst->getParent()->getParent() == CurrFn) {
							// This use of V2 belongs to this function -- Find the nonConst version of it, and instrument it
							Instruction *nonConstI2;
							Value *nonConstV2;
							for(Function::iterator fit = CurrFn->begin(); fit != CurrFn->end(); fit++) {
								if(fit == V2UseInst->getParent()) {
									for(BasicBlock::iterator bbit = fit->begin(); bbit != fit->end(); bbit++) {
										if(bbit == V2UseInst) {
											nonConstI2 = bbit;
											// Find nonConstV2
											#if 0
											if(LoadInst* LI = dyn_cast<LoadInst>(bbit)) {
												assert(V2 == LI->getOperand(0));
												nonConstV2 = bbit->getOperand(0);
											}
											else if(StoreInst* SI = dyn_cast<StoreInst>(bbit)) {
												assert(V2 == SI->getOperand(1));
												nonConstV2 = SI->getOperand(1);
											}
											else {
											#endif
												//outs() << "bbit:(!Load !Store): " << *bbit << "\n";
												// Check the other operands to find the nonConst version of V2
												for(unsigned int op_index = 0; op_index < nonConstI2->getNumOperands(); 
													op_index++) {
													if(nonConstI2->getOperand(op_index) == V2) {
														nonConstV2 = nonConstI2->getOperand(op_index);
														break;
													}
												}
											#if 0
											}
											#endif
										}
									}
								}
							}
							assert(nonConstI2 != NULL);
							assert(nonConstV2 != NULL);

							//Instrument this nonConstV1
							SubprimeArgs[0] = nonConstV2;
							SubprimeArgs[1] = ConstantInt::get(Type::getInt32Ty(*CurrContext), CurrPrimeID, false);

					 		Instruction *PinSubprimeInsertInst = 
								CallInst::Create(PinSubprimeInsertFn, SubprimeArgs.begin(), SubprimeArgs.end(),"", nonConstI2);
						}
					}
				}
			}
		}
#ifndef ADD_TO_SPECLIST
		assert(count == 0);
#endif
		outs() << "Queries: " << count << "\n";
#endif


	}

}

ModulePass* llvm::createFuncClonePass() { return new FuncClone(); }
